# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - ImprovedMoinSearch library

    @copyright: 2010 Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.search import searchPages
from MoinMoin.Page import Page
from MoinMoin import wikiutil
import re

heading_regexp = re.compile(r"^(?P<level>=+)(?P<heading>.*?)(?P=level)$", re.UNICODE | re.MULTILINE)
paragraph_regexp = re.compile(r"(?P<paragraph>(?:^[^#=\s].*$\n)+)", re.UNICODE | re.MULTILINE)

def range_groups(min_name, max_name):
    return r"(?P<%s>-?\d+)?(?:\s*-\s*(?P<%s>-?\d+))?" % (min_name, max_name)

format_options_regexp = re.compile(
    r"(?P<link>\[)?"
    r"("
    r"(?P<heading>(heading|title|h)\s*" + range_groups("min_heading", "max_heading") + ")"
    r"|(?P<paragraph>(paragraph|para|p)\s*(?P<paragraph_number>\d+)?)"
    r"|(?P<name>(name|page)\s*" + range_groups("first", "last") + ")"
    r")"
    r"(?(link)\]|)",
    re.UNICODE)

def convert_index(i, length):

    """
    Convert from a 1-based indexing scheme to a 0-based scheme for the given
    index 'i' in a sequence having the given 'length'.
    """

    if i is None:
        return i
    elif i > 0:
        return i - 1
    elif i < 0:
        return length + i
    else:
        return i

def getSearchResultPages(request, query, **kw):

    """
    Return matching pages using the given 'request' and search 'query'. Optional
    keyword arguments are passed to the underlying search infrastructure.
    """

    results = searchPages(request, query, **kw)
    return results.hits

def getFirstPageHeading(request, page, start=0, min_level=None, max_level=None):

    """
    Using the given 'request', return the first heading in the given 'page'
    from the given 'start' point (optional, defaulting to the start of the page)
    having a heading level of at least 'min_level' (which is undefined if not
    specified) and at most 'max_level' (which is undefined if not specified).

    A tuple containing the heading and the span (the start offset and the end
    offset as a tuple) is returned for a successful retrieval. Otherwise, None
    is returned.
    """

    full_page = Page(request, page.page_name)
    body = full_page.get_raw_body()
    if start != 0:
        body = body[start:]

    for match in heading_regexp.finditer(body):
        level = len(match.group("level"))

        if (min_level is None or level >= min_level) and \
            (max_level is None or level <= max_level):

            return match.group("heading"), match.span()

    return None

def getParagraph(request, page, start=0, number=None):

    """
    Using the given 'request', return from the given 'page', starting from the
    optional 'start' offset (or the beginning, if no such offset is specified),
    the first paragraph or, if the optional 'number' is given, the paragraph
    whose position corresponds to that number, with a number of 1 being the
    first paragraph found, 2 being the second, and so on.
    """

    full_page = Page(request, page.page_name)
    body = full_page.get_raw_body()
    if start != 0:
        body = body[start:]

    for i, match in enumerate(paragraph_regexp.finditer(body)):
        if number is None or i == max(0, number - 1):
            return match.group("paragraph"), match.span()

    return None

def getPageName(request, page, start=0, first=None, last=None):

    """
    Using the given 'request', return the name of the given 'page'. The optional
    'start' offset refers to the body of the page and is returned as the start
    and end of the result span if specified.

    If the optional 'first' or 'last' parameters are specified, only the
    specified span of parts extracted from the page name will be returned, where
    the parts of the name are obtained by splitting the full name where the
    slash ("/") character is found. The first part has an index of 1, and the
    last part can be referred to using an index of -1.
    """

    parts = page.page_name.split("/")

    first = convert_index(first, len(parts))
    last = convert_index(last, len(parts))

    if first is None:
        if last is None:
            pass
        else:
            parts = parts[:last+1]
    else:
        if last is None:
            parts = parts[first:]
        else:
            parts = parts[first:last+1]

    return "/".join(parts), (start, start)

def formatResultPages(request, formatter, pages, paging, format, page_from=0):

    """
    Using the given 'request' and 'formatter', return a formatted string showing
    the result 'pages', providing paging controls when 'paging' is set to a true
    value, and providing page details according to the given 'format'.

    If the optional 'pages_from' parameter is set, the result pages from the
    given result (specified within a range from 0 to the length of the 'pages'
    collection) will be shown.
    """

    actions = []

    if format:
        for match in format_options_regexp.finditer(format):
            as_link = match.group("link")
            if match.group("heading"):
                actions.append((getFirstPageHeading, map(int_or_none, (match.group("min_heading"), match.group("max_heading"))), as_link))
            elif match.group("paragraph"):
                actions.append((getParagraph, map(int_or_none, (match.group("paragraph_number"),)), as_link))
            elif match.group("name"):
                actions.append((getPageName, map(int_or_none, (match.group("first"), match.group("last"))), as_link))
    else:
        actions.append((getPageName, (), True))

    # Use paging only when there are enough results.

    results_per_page = request.cfg.search_results_per_page
    paging = paging and len(pages) > results_per_page

    if paging:
        pages_to_show = pages[page_from:page_from + results_per_page]
    else:
        pages_to_show = pages

    # Prepare the output.

    output = []
    output.append(formatter.number_list(on=1, start=page_from + 1))

    for page in pages_to_show:
        output.append(formatter.listitem(on=1))

        start = 0
        first = 1
        for action, args, as_link in actions:
            result = action(request, page, start, *args)

            if result is not None:
                if not first:
                    output.append(" ")
                if as_link:
                    output.append(formatter.pagelink(on=1, pagename=page.page_name))

                text, span = result
                output.append(formatter.text(text))

                # Position the search for the next action.

                _start, _end = span
                start = _end + 1

                if as_link:
                    output.append(formatter.pagelink(on=0))

            first = 0

        output.append(formatter.listitem(on=0))

    output.append(formatter.number_list(on=0))

    # Show paging navigation.

    if paging:
        output.append(formatPagingNavigation(request, formatter, pages, page_from))

    return "".join(output)

def formatPagingNavigation(request, formatter, pages, page_from=0):

    """
    Using the given 'request' and 'formatter', return a formatted string showing
    the paging navigation for the result 'pages', according to the 'page_from'
    indicator which provides the current position in the result set.
    """

    page = formatter.page
    pagename = page.page_name
    _ = request.getText

    output = []

    results_per_page = request.cfg.search_results_per_page
    number_of_results = len(pages)

    pages_total = number_of_results / results_per_page
    pages_before = page_from / results_per_page
    pages_after = ((number_of_results - page_from) / results_per_page) - 1

    querydict = wikiutil.parseQueryString(request.query_string)

    output.append(formatter.paragraph(on=1))
    output.append(formatter.text(_("Result pages:")))
    output.append(formatter.text(" "))

    n = 0
    while n < pages_before:
        output.append(formatter.pagelink(on=1, pagename=pagename, querystr=getPagingQueryString(querydict, n * results_per_page)))
        output.append(formatter.text(str(n + 1)))
        output.append(formatter.pagelink(on=0))
        output.append(formatter.text(" "))
        n += 1

    output.append(formatter.text(str(n + 1)))
    output.append(formatter.text(" "))
    n += 1

    while n <= pages_total:
        output.append(formatter.pagelink(on=1, pagename=pagename, querystr=getPagingQueryString(querydict, n * results_per_page)))
        output.append(formatter.text(str(n + 1)))
        output.append(formatter.pagelink(on=0))
        output.append(formatter.text(" "))
        n += 1

    output.append(formatter.paragraph(on=0))

    return "".join(output)

def getPagingQueryString(querydict, page_from):
    querydict["from"] = page_from
    return wikiutil.makeQueryString(querydict)

def int_or_none(x):
    if x is None:
        return x
    else:
        return int(x)

# vim: tabstop=4 expandtab shiftwidth=4
